#pragma once

//#include "../myincludes.hh"

#include<dune/pdelab/common/quadraturerules.hh>
#include<dune/pdelab/finiteelement/localbasiscache.hh>
//#include<dune/pdelab/localoperator/idefault.hh>

//Parameter class (for functions), FEM space and RangeField
template<typename Param, typename CHGFS, typename CHContainer, typename NSGFS, typename NSContainer>
class CHNS_FEM_CahnHilliard :
  public Dune::PDELab::FullVolumePattern,
  public Dune::PDELab::LocalOperatorDefaultFlags,
  public Dune::PDELab::NumericalJacobianVolume<CHNS_FEM_CahnHilliard<Param, CHGFS, CHContainer, NSGFS, NSContainer> >,
  public Dune::PDELab::NumericalJacobianApplyVolume<CHNS_FEM_CahnHilliard<Param, CHGFS, CHContainer, NSGFS, NSContainer> >
{
private:
  typedef typename CHContainer::field_type RF;

  typedef typename CHGFS::template Child<0>::Type FEM;
  typedef typename FEM::Traits::FiniteElementType::Traits::LocalBasisType LocalBasis;
  Dune::PDELab::LocalBasisCache<LocalBasis> cache;

  typedef typename NSGFS::template Child<0>::Type::template Child<0>::Type VFEM;
  typedef typename VFEM::Traits::FiniteElementType::Traits::LocalBasisType VLocalBasis;
  Dune::PDELab::LocalBasisCache<VLocalBasis> vcache;

  Param& param; // parameter functions

  //FOR EVALUATING the old cahn-Hilliard values
  typedef Dune::PDELab::LocalFunctionSpace<CHGFS> CHLFS;
  mutable CHLFS chLfs;
  typedef Dune::PDELab::LFSIndexCache<CHLFS> CHLFSCache;
  mutable CHLFSCache chLfsCache;
  typedef typename CHContainer::template ConstLocalView<CHLFSCache> CHView;
  mutable CHView chView;
  mutable std::vector<typename LocalBasis::Traits::RangeFieldType> chOldLocal;

  //FOR EVALUATING the old navier-Stokes values
  typedef Dune::PDELab::LocalFunctionSpace<NSGFS> NSLFS;
  mutable NSLFS nsLfs;
  typedef Dune::PDELab::LFSIndexCache<NSLFS> NSLFSCache;
  mutable NSLFSCache nsLfsCache;
  typedef typename NSContainer::template ConstLocalView<NSLFSCache> NSView;
  mutable NSView nsView;
  mutable std::vector<typename VLocalBasis::Traits::RangeFieldType> nsOldLocal;

  //FOR DETECTING ENTITY changes
  typedef typename NSGFS::Traits::GridViewType::template Codim<0>::Entity   ENT;
  mutable ENT OldEnt;

public:
  enum {doPatternVolume=true};
  //enum {doLambdaVolume=true};
  enum {doAlphaVolume=true};

  //constructor
  CHNS_FEM_CahnHilliard(Param& param_, CHGFS& chGfs_, CHContainer& chVOld_, NSGFS& nsGfs_, NSContainer& nsVOld_) :
    param(param_),
    chLfs(chGfs_), chLfsCache(chLfs), chView(chVOld_), chOldLocal(chGfs_.maxLocalSize()),
    nsLfs(nsGfs_), nsLfsCache(nsLfs), nsView(nsVOld_), nsOldLocal(nsGfs_.maxLocalSize())
  {
  }


  template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
  void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
  {
    //Left hand side, terms depending on the solution x
    using namespace Dune::Indices;

    //Quadrature Rule
    const int dim = EG::Entity::dimension;
    const int order = 2; //TODO: Choose more educated
    auto rule = Dune::PDELab::quadratureRule(eg.geometry(), order);
    //const auto& myordering = lfsv.gridFunctionSpace().ordering();

    const auto& pfspace = child(lfsv,_0);
    const auto& xispace = child(lfsv,_1);
    const auto& femspace = child(lfsv,_0);

    const auto& vspace = child(nsLfs,_0);
    const auto& vfemspace = child(vspace,_0);

    if(eg.entity() != OldEnt)
    {
      OldEnt = eg.entity();

      chLfs.bind(eg.entity());
      chLfsCache.update();
      chView.bind(chLfsCache);
      chView.read(chOldLocal);
      chView.unbind();

      nsLfs.bind(eg.entity());
      nsLfsCache.update();
      nsView.bind(nsLfsCache);

      nsView.read_sub_container(vspace,nsOldLocal);
      nsView.unbind();
    }

    for(const auto& ip : rule)
    {
      // evaluate shape functions and  gradient of shape functions
      auto& phi = cache.evaluateFunction(ip.position(), femspace.finiteElement().localBasis());
      auto& gradphihat = cache.evaluateJacobian(ip.position(), femspace.finiteElement().localBasis());

      auto& vphi = vcache.evaluateFunction(ip.position(), vfemspace.finiteElement().localBasis());

      // transform gradients of shape functions to real element
      const auto S = eg.geometry().jacobianInverseTransposed(ip.position());
      auto gradphi = makeJacobianContainer(femspace);
      for (size_t i=0; i<femspace.size(); i++)
        S.mv(gradphihat[i][0],gradphi[i][0]);

      //compute current values of the unknowns
      Dune::FieldVector<RF,dim> v(0.0);
      RF pf=0.0;
      RF pfOld=0.0;
      RF xi=0.0;
      Dune::FieldVector<RF,dim> gradpf(0.0);
      Dune::FieldVector<RF,dim> gradxi(0.0);

      for (size_t i=0; i<femspace.size(); i++)
      {

        gradpf.axpy(x(pfspace,i),gradphi[i][0]);
        pf += x(pfspace,i)*phi[i];

        gradxi.axpy(x(xispace,i),gradphi[i][0]);
        xi += x(xispace,i)*phi[i];

        pfOld += chOldLocal[pfspace.localIndex(i)]*phi[i];

      }
      for (size_t i=0; i<vfemspace.size(); i++)
      {
        for(size_t k=0; k<dim; k++)
        {
          v[k] += nsOldLocal[child(vspace,k).localIndex(i)]*vphi[i];
        }
      }

      auto factor = ip.weight()*eg.geometry().integrationElement(ip.position());
      for (size_t i=0; i<femspace.size(); i++)
      {
      // integrate: (pf-pfold)*phi_i/delta_t + gradpf * grad phi_i + ...
        r.accumulate(pfspace,i,factor*(
          (pf-pfOld)*phi[i]/param.dt
          +param.Mob(pfOld)*param.eps* (gradxi*gradphi[i][0])
          -pf*(v*gradphi[i][0])
          ));

        r.accumulate(xispace,i,factor*(
          xi*phi[i]
          - 1/param.eps* (param.dWConvex(pf)+param.dWConcarve(pfOld))*phi[i]
          - param.eps/2 * (gradpf*gradphi[i][0])
        ));
      }
    }
  }

};
