#pragma once

#include "../myincludes.hh"

#include <dune/istl/io.hh>
#include <dune/istl/operators.hh>
#include <dune/istl/owneroverlapcopy.hh>
#include <dune/istl/paamg/amg.hh>
#include <dune/istl/paamg/pinfo.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/scalarproducts.hh>
#include <dune/istl/solvercategory.hh>
#include <dune/istl/solvers.hh>
#include <dune/istl/superlu.hh>

template<class GO,
         template<class,class,class,int> class Preconditioner,
         template<class, class, class> class Solver>
class ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES
{
  typedef typename GO::Traits::TrialGridFunctionSpace GFS;
  typedef Dune::PDELab::ISTL::ParallelHelper<GFS> PHELPER;

public:
  explicit ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES (const GO& grid_operator,
    unsigned maxiter_ = 5000, unsigned steps_ = 5, int verbose_ = 1)
    : _grid_operator(grid_operator)
    , gfs(grid_operator.trialGridFunctionSpace())
    , phelper(gfs,verbose_)
    , maxiter(maxiter_)
    , steps(steps_)
    , verbose(verbose_)
  {}

  template<class Vector>
  typename Vector::ElementType norm (const Vector& v) const
  {
    Vector x(v); // make a copy because it has to be made consistent
    typedef Dune::PDELab::NonoverlappingScalarProduct<GFS,Vector> PSP;
    PSP psp(gfs,phelper);
    psp.make_consistent(x);
    return psp.norm(x);
  }

  template<class M, class V, class W>
  void apply(M& A, V& z, W& r, typename V::ElementType reduction)
  {
    using MatrixType = Dune::PDELab::Backend::Native<M>;
    MatrixType& mat = Dune::PDELab::Backend::native(A);
    using VectorType = Dune::PDELab::Backend::Native<W>;
#if HAVE_MPI
    typedef typename Dune::PDELab::ISTL::CommSelector<96,Dune::MPIHelper::isFake>::type Comm;
    _grid_operator.make_consistent(A);
    Dune::PDELab::ISTL::assertParallelUG(gfs.gridView().comm());
    Comm oocc(gfs.gridView().comm(),Dune::SolverCategory::nonoverlapping);
    phelper.createIndexSetAndProjectForAMG(mat, oocc);
    typedef Preconditioner<MatrixType,VectorType,VectorType,1> Smoother;
    Smoother smoother(mat, /*steps,*/ 1.0);
    typedef Dune::NonoverlappingSchwarzScalarProduct<VectorType,Comm> PSP;
    PSP psp(oocc);
    typedef Dune::NonoverlappingSchwarzOperator<MatrixType,VectorType,VectorType,Comm> Operator;
    Operator oop(mat,oocc);
    typedef Dune::NonoverlappingBlockPreconditioner<Comm, Smoother> ParSmoother;
    ParSmoother parsmoother(smoother, oocc);
#else
    typedef Preconditioner<MatrixType,VectorType,VectorType,1> ParSmoother;
    ParSmoother parsmoother(mat, steps, 1.0);
    typedef Dune::SeqScalarProduct<VectorType> PSP;
    PSP psp;
    typedef Dune::MatrixAdapter<MatrixType,VectorType,VectorType> Operator;
    Operator oop(mat);
#endif
    int verb=0;
    if (gfs.gridView().comm().rank()==0) verb=verbose;
    Solver<VectorType, VectorType, VectorType> solver(oop,psp,parsmoother,reduction,100,maxiter,verb);
    Dune::InverseOperatorResult stat;
    //make r consistent
    if (gfs.gridView().comm().size()>1){
      Dune::PDELab::AddDataHandle<GFS,V> adddh(gfs,r);
      gfs.gridView().communicate(adddh,
                                 Dune::InteriorBorder_InteriorBorder_Interface,
                                 Dune::ForwardCommunication);
    }

    solver.apply(z,r,stat);
    res.converged  = stat.converged;
    res.iterations = stat.iterations;
    res.elapsed    = stat.elapsed;
    res.reduction  = stat.reduction;
    res.conv_rate  = stat.conv_rate;
  }

  const Dune::PDELab::LinearSolverResult<double>& result() const
  {
    return res;
  }

private:
  const GO& _grid_operator;
  const GFS& gfs;
  PHELPER phelper;
  Dune::PDELab::LinearSolverResult<double> res;
  unsigned maxiter;
  unsigned steps;
  int verbose;
};
